"""Tests the STACS pack model and validator."""

import json
import os
import unittest

import stacs.scan


class STACSModelPackTestCase(unittest.TestCase):
    """Tests the STACS pack model and validator."""

    def setUp(self):
        """Ensure the application is setup for testing."""
        self.fixtures_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "fixtures/pack/"
        )

    def tearDown(self):
        """Ensure everything is torn down between tests."""
        pass

    def test_simple_pack(self):
        """Ensure that simple packs can be loaded."""
        with open(os.path.join(self.fixtures_path, "001-simple.valid.json"), "r") as f:
            stacs.scan.model.pack.Format(**json.load(f))



"""Tests the STACS allow list model and validator."""

import json
import os
import unittest

import stacs.scan


class STACSModelAllowListTestCase(unittest.TestCase):
    """Tests the STACS allow list model and validator."""

    def setUp(self):
        """Ensure the application is setup for testing."""
        self.fixtures_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "fixtures/ignore_list/"
        )

    def tearDown(self):
        """Ensure everything is torn down between tests."""
        pass

    def test_simple(self):
        """Ensure that simple allow lists can be loaded."""
        with open(os.path.join(self.fixtures_path, "001-simple.valid.json"), "r") as f:
            stacs.scan.model.ignore_list.Format(**json.load(f))

    def test_hierarchical_loading(self):
        """Ensure that hierarchical allow lists can be loaded."""
        with open(os.path.join(self.fixtures_path, "002-project.valid.json"), "r") as f:
            stacs.scan.model.ignore_list.Format(**json.load(f))



"""Define constants commonly used throughout STACS.

SPDX-License-Identifier: BSD-3-Clause
"""

# The size of chunks to use when reading files.
CHUNK_SIZE = 65536

# The size, in bytes, of the sample window.
WINDOW_SIZE = 20

# Define the default cache directory, used to unpack archives into.
CACHE_DIRECTORY = "/tmp"

# Define the character to use when constructed paths to findings which are inside of
# archives.
ARCHIVE_FILE_SEPARATOR = "!"

# Define an exit code to use when there are unsuppressed findings.
EXIT_CODE_UNSUPPRESSED = 100

# External licenses will be displayed during STACS banner.
EXTERNAL_LICENSES = {
    "libarchive": [
        "https://github.com/libarchive/libarchive/blob/master/COPYING",
    ],
    "yara": [
        "https://github.com/VirusTotal/yara-python/blob/master/LICENSE",
    ],
}



"""Tests the STACS ignore list filter."""

import os
import unittest

import stacs.scan


class STACSFilterIgnoreListTestCase(unittest.TestCase):
    """Tests the STACS ignore list filter."""

    def setUp(self):
        """Ensure the application is setup for testing."""
        self.fixtures_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "fixtures/ignore_list/"
        )

    def tearDown(self):
        """Ensure everything is torn down between tests."""
        pass

    def test_by_path(self):
        """Validate whether path filters are working."""
        # Use the same fixture for all branches.
        finding = stacs.scan.model.finding.Entry(
            path="/a/a",
            md5="fa19207ef28b6a97828e3a22b11290e9",
            location=stacs.scan.model.finding.Location(
                offset=300,
            ),
            source=stacs.scan.model.finding.Source(
                module="stacs.scan.scanner.rules",
                reference="SomeRule",
            ),
        )

        # Define ignores which should correctly be ignored.
        hits = [
            # Path matches, no other constraint.
            stacs.scan.model.ignore_list.Entry(path="/a/a", reason="Test"),
            # Path matches, reference matches.
            stacs.scan.model.ignore_list.Entry(
                path="/a/a", reason="Test", references=["SomeRule", "OtherRule"]
            ),
            # Path matches, offset matches.
            stacs.scan.model.ignore_list.Entry(path="/a/a", reason="Test", offset=300),
        ]

        # Path differs.
        miss = stacs.scan.model.ignore_list.Entry(path="/a/b", reason="Test")
        self.assertEqual(stacs.scan.filter.ignore_list.by_path(finding, miss), False)

        # Path matches, reference differs.
        miss = stacs.scan.model.ignore_list.Entry(
            path="/a/a", reason="Test", references=["OtherRule"]
        )
        self.assertEqual(stacs.scan.filter.ignore_list.by_path(finding, miss), False)

        # Path matches, offset differs.
        miss = stacs.scan.model.ignore_list.Entry(
            path="/a/a", reason="Test", offset=1234
        )
        self.assertEqual(stacs.scan.filter.ignore_list.by_path(finding, miss), False)

        # Ensure all hit entries are matches.
        for hit in hits:
            self.assertEqual(stacs.scan.filter.ignore_list.by_path(finding, hit), True)

    def test_by_pattern(self):
        """Validate whether pattern filters are working."""
        # Use the same fixture for all branches.
        finding = stacs.scan.model.finding.Entry(
            path="/a/tests/a",
            md5="fa19207ef28b6a97828e3a22b11290e9",
            location=stacs.scan.model.finding.Location(
                offset=300,
            ),
            source=stacs.scan.model.finding.Source(
                module="stacs.scan.scanner.rules",
                reference="SomeRule",
            ),
        )

        # Pattern matches, no other constraint.
        hit = stacs.scan.model.ignore_list.Entry(pattern=".*/tests/.*", reason="Test")
        self.assertEqual(stacs.scan.filter.ignore_list.by_pattern(finding, hit), True)

        # Pattern matches, reference matches.
        hit = stacs.scan.model.ignore_list.Entry(
            pattern=".*/tests/.*",
            reason="Test",
            references=["SomeRule", "OtherRule"],
        )
        self.assertEqual(stacs.scan.filter.ignore_list.by_pattern(finding, hit), True)

        # Pattern matches, offset matches.
        hit = stacs.scan.model.ignore_list.Entry(
            pattern=".*/tests/.*", reason="Test", offset=300
        )
        self.assertEqual(stacs.scan.filter.ignore_list.by_pattern(finding, hit), True)

        # Pattern differs.
        miss = stacs.scan.model.ignore_list.Entry(pattern=r"\.shasums$", reason="Test")
        self.assertEqual(stacs.scan.filter.ignore_list.by_pattern(finding, miss), False)

        # Pattern matches, reference differs.
        miss = stacs.scan.model.ignore_list.Entry(
            pattern=".*/tests/.*", reason="Test", references=["OtherRule"]
        )
        self.assertEqual(stacs.scan.filter.ignore_list.by_pattern(finding, miss), False)

        # Pattern matches, offset differs.
        miss = stacs.scan.model.ignore_list.Entry(
            pattern=".*/tests/.*", reason="Test", offset=1234
        )
        self.assertEqual(stacs.scan.filter.ignore_list.by_pattern(finding, miss), False)

    def test_by_hash(self):
        """Validate whether hash filters are working."""
        # Use the same fixture for all branches.
        finding = stacs.scan.model.finding.Entry(
            path="/a/tests/a",
            md5="fa19207ef28b6a97828e3a22b11290e9",
            location=stacs.scan.model.finding.Location(
                offset=300,
            ),
            source=stacs.scan.model.finding.Source(
                module="stacs.scan.scanner.rules",
                reference="SomeRule",
            ),
        )

        # Hash matches, no other constraint.
        hit = stacs.scan.model.ignore_list.Entry(
            md5="fa19207ef28b6a97828e3a22b11290e9", reason="Test"
        )
        self.assertEqual(stacs.scan.filter.ignore_list.by_hash(finding, hit), True)

        # Hash matches, reference matches.
        hit = stacs.scan.model.ignore_list.Entry(
            md5="fa19207ef28b6a97828e3a22b11290e9",
            reason="Test",
            references=["SomeRule", "OtherRule"],
        )
        self.assertEqual(stacs.scan.filter.ignore_list.by_hash(finding, hit), True)

        # Hash matches, offset matches.
        hit = stacs.scan.model.ignore_list.Entry(
            md5="fa19207ef28b6a97828e3a22b11290e9", reason="Test", offset=300
        )
        self.assertEqual(stacs.scan.filter.ignore_list.by_hash(finding, hit), True)

        # Hash differs.
        miss = stacs.scan.model.ignore_list.Entry(
            md5="cf42e6f36da80658591489975bbd845b", reason="Test"
        )
        self.assertEqual(stacs.scan.filter.ignore_list.by_hash(finding, miss), False)

        # Hash matches, reference differs.
        miss = stacs.scan.model.ignore_list.Entry(
            md5="fa19207ef28b6a97828e3a22b11290e9",
            reason="Test",
            references=["OtherRule"],
        )
        self.assertEqual(stacs.scan.filter.ignore_list.by_hash(finding, miss), False)

        # Hash matches, offset differs.
        miss = stacs.scan.model.ignore_list.Entry(
            md5="fa19207ef28b6a97828e3a22b11290e9", reason="Test", offset=1234
        )
        self.assertEqual(stacs.scan.filter.ignore_list.by_hash(finding, miss), False)



"""Defines outputs supported by STACS.

SPDX-License-Identifier: BSD-3-Clause
"""

from stacs.scan.output import markdown  # noqa: F401
from stacs.scan.output import pretty  # noqa: F401
from stacs.scan.output import sarif  # noqa: F401



"""Defines handlers for unpacking of archives.

SPDX-License-Identifier: BSD-3-Clause
"""

import bz2
import gzip
import hashlib
import logging
import lzma
import os
import shutil
import tarfile
import zipfile
import zlib
from typing import List, Tuple

import zstandard
from stacs.native import archive
from stacs.scan.constants import CHUNK_SIZE
from stacs.scan.exceptions import FileAccessException, InvalidFileException
from stacs.scan.loader.format import dmg, xar


def path_hash(filepath: str) -> str:
    """Returns a hash of the filepath, for use with unique directory creation."""
    return hashlib.md5(bytes(filepath, "utf-8")).hexdigest()


def zip_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided zip archive."""
    log = logging.getLogger(__name__)

    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    # Attempt to unpack the zipfile to the new unpack directory.
    try:
        with zipfile.ZipFile(filepath, "r") as reader:
            try:
                reader.extractall(directory)
            except RuntimeError as err:
                # Encrypted zips (why is this not a custom exception?!)
                if "encrypted" in str(err):
                    log.warn(
                        f"Cannot process file in archive at {filepath}, skipping: {err}"
                    )
            except NotADirectoryError as err:
                # Broken filepaths inside of ZIP.
                log.warn(
                    f"Cannot process file in archive at {filepath}, skipping: {err}"
                )
            except (OSError, IndexError) as err:
                # Several conditions, but usually a corrupt / bad input zip.
                log.warn(
                    f"Cannot process file in archive at {filepath}, skipping: {err}"
                )
    except (zipfile.BadZipFile, OSError) as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {directory}: {err}"
        )


def tar_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided tarball."""
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    # Attempt to unpack the tarball to the new unpack directory.
    try:
        with tarfile.open(filepath, "r") as reader:
            reader.extractall(directory)
    except (PermissionError, tarfile.TarError) as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {directory}: {err}"
        )


def gzip_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided gzip archive."""
    output = ".".join(os.path.basename(filepath).split(".")[:-1])

    # No dots? Just use the name as is.
    if len(output) < 1:
        output = os.path.basename(filepath)

    # Although gzip files cannot contain more than one file, we'll still spool into
    # a subdirectory under the cache for consistency.
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    # TODO: This can likely be optimized for tgz files, as currently the file will be
    #       first processed and gunzipped, and then reprocessed to be extracted.
    try:
        with gzip.open(filepath, "rb") as fin:
            with open(os.path.join(directory, output), "wb") as fout:
                shutil.copyfileobj(fin, fout, CHUNK_SIZE)
    except gzip.BadGzipFile as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {output}: {err}"
        )


def bzip2_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided bzip2 archive."""
    output = ".".join(os.path.basename(filepath).split(".")[:-1])

    # No dots? Just use the name as is.
    if len(output) < 1:
        output = os.path.basename(filepath)

    # Like gzip, bzip2 cannot support more than a single file. Again, we'll spool into
    # a subdirectory for consistency.
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    # TODO: This can likely be optimized for tbz files, as currently the file will be
    #       first processed and gunzipped, and then reprocessed to be extracted.
    try:
        with bz2.open(filepath, "rb") as fin:
            with open(os.path.join(directory, output), "wb") as fout:
                shutil.copyfileobj(fin, fout, CHUNK_SIZE)
    except (OSError, ValueError) as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {output}: {err}"
        )


def zstd_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided zstd archive."""
    output = ".".join(os.path.basename(filepath).split(".")[:-1])

    # No dots? Just use the name as is.
    if len(output) < 1:
        output = os.path.basename(filepath)

    # zstd does not appear to provide a native mechanism to compress multiple files,
    # and recommend 'to combine zstd with tar'.
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    try:
        decompressor = zstandard.ZstdDecompressor()

        with open(filepath, "rb") as fin:
            with open(os.path.join(directory, output), "wb") as fout:
                decompressor.copy_stream(fin, fout, read_size=CHUNK_SIZE)
    except (OSError, ValueError, zstandard.ZstdError) as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {output}: {err}"
        )


def lzma_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided xz / lzma archive."""
    output = ".".join(os.path.basename(filepath).split(".")[:-1])

    # No dots? Just use the name as is.
    if len(output) < 1:
        output = os.path.basename(filepath)

    # Although xz files cannot contain more than one file, we'll still spool into
    # a subdirectory under the cache for consistency.
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    try:
        with lzma.open(filepath, "rb") as fin:
            with open(os.path.join(directory, output), "wb") as fout:
                shutil.copyfileobj(fin, fout, CHUNK_SIZE)
    except lzma.LZMAError as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {output}: {err}"
        )


def zlib_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided zlib archive."""
    output = ".".join(os.path.basename(filepath).split(".")[:-1])

    # No dots? Just use the name as is.
    if len(output) < 1:
        output = os.path.basename(filepath)

    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    try:
        decompressor = zlib.decompressobj(wbits=zlib.MAX_WBITS)

        with open(filepath, "rb") as fin:
            with open(os.path.join(directory, output), "wb") as fout:
                while compressed := fin.read(CHUNK_SIZE):
                    fout.write(decompressor.decompress(compressed))
    except zlib.error as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {output}: {err}"
        )


def xar_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided XAR archive."""
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    # Attempt to unpack the archive.
    try:
        archive = xar.XAR(filepath)
        archive.extract(directory)
    except FileAccessException as err:
        raise FileAccessException(
            f"Unable to extract archive {filepath} to {directory}: {err}"
        )
    except InvalidFileException as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {directory}: {err}"
        )


def dmg_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided DMG archive."""
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    # Attempt to unpack the archive.
    try:
        archive = dmg.DMG(filepath)
        archive.extract(directory)
    except FileAccessException as err:
        raise FileAccessException(
            f"Unable to extract archive {filepath} to {directory}: {err}"
        )
    except InvalidFileException as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {directory}: {err}"
        )


def libarchive_handler(filepath: str, directory: str) -> None:
    """Attempts to extract the provided archive with libarchive."""
    try:
        os.mkdir(directory, mode=0o700)
    except OSError as err:
        raise FileAccessException(
            f"Unable to create unpack directory at {directory}: {err}"
        )

    # Attempt to unpack the archive to the new unpack directory.
    try:
        with archive.ArchiveReader(filepath) as reader:
            for entry in reader:
                member = entry.filename
                member = member.lstrip("../")
                member = member.lstrip("./")

                if entry.filename == ".":
                    continue

                destination = os.path.join(directory, member)
                parent = os.path.dirname(destination)

                # Handle odd cases where a file was created where a directory needs to
                # be.
                if os.path.exists(parent) and os.path.isfile(parent):
                    os.unlink(parent)

                if os.path.isdir(destination):
                    continue

                # Create parent directories, as required.
                if not os.path.isdir(parent):
                    os.makedirs(parent)

                # If the entry is a directory, create it and move on.
                if entry.isdir:
                    os.makedirs(destination, exist_ok=True)
                    continue

                with open(destination, "wb") as fout:
                    while True:
                        chunk = reader.read()
                        if len(chunk) > 0:
                            fout.write(chunk)
                            continue
                        break
    except archive.ArchiveError as err:
        raise InvalidFileException(
            f"Unable to extract archive {filepath} to {directory}: {err}"
        )


def get_mimetype(chunk: bytes, start: bool) -> List[Tuple[int, str]]:
    """Attempts to locate the appropriate handler for a given file.

    This may fail if the required "magic" is at an offset greater than the CHUNK_SIZE.
    However, currently this is not an issue, but may need to be revisited later as more
    archive types are supported.

    The start flag is used to indicate whether the current chunk is from the start of
    the file, or the end of the file. Today we only support checking the first and last
    chunk.

    Returns a list of weights and MIME types as a tuple. This weight is specified by
    handlers and is used to allow "container" formats, which may contain multiple other
    files of various matching types, to "win" the match - due to a higher weight.
    """
    for name, options in MIME_TYPE_HANDLERS.items():
        offset = options["offset"]
        magic = options["magic"]

        # If looking at the last chunk, only use negative offsets. This is to prevent
        # false positives as position 0 in the last chunk is actually N bytes into the
        # file. This is especially problematic for formats with short magic numbers,
        # such as zlib.
        if not start and offset >= 0:
            continue

        # TODO: How to handle multiple matches in the same chunk? Is this this likely?
        for format in magic:
            if chunk[offset : (offset + len(format))] == format:  # noqa: E203
                return (options["weight"], name)

    return (0, None)


# Define all supported archives and their handlers. As we currently only support a small
# list of types we can just define file magic directly here, rather than use an external
# library. This removes the need for dependencies which may have other system
# dependencies - such as libmagic. It should also provide a small a speed up during
# unpacking, as we're only looking for a small number of types.
MIME_TYPE_HANDLERS = {
    "application/x-tar": {
        "weight": 1,
        "offset": 257,
        "magic": [
            bytearray([0x75, 0x73, 0x74, 0x61, 0x72]),
        ],
        "handler": tar_handler,
    },
    "application/gzip": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0x1F, 0x8B]),
        ],
        "handler": gzip_handler,
    },
    "application/x-bzip2": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0x42, 0x5A, 0x68]),
        ],
        "handler": bzip2_handler,
    },
    "application/zip": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0x50, 0x4B, 0x03, 0x04]),
            bytearray([0x50, 0x4B, 0x05, 0x06]),
            bytearray([0x50, 0x4B, 0x07, 0x08]),
        ],
        "handler": zip_handler,
    },
    "application/zlib": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0x78, 0x01]),
            bytearray([0x78, 0x5E]),
            bytearray([0x78, 0x9C]),
            bytearray([0x78, 0xDA]),
        ],
        "handler": zlib_handler,
    },
    "application/x-xz": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0xFD, 0x37, 0x7A, 0x58, 0x5A, 0x00]),
        ],
        "handler": lzma_handler,
    },
    "application/x-rpm": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0xED, 0xAB, 0xEE, 0xDB]),
        ],
        "handler": libarchive_handler,
    },
    "application/x-iso9660-image": {
        "weight": 1,
        "offset": 0x8001,
        "magic": [
            bytearray([0x43, 0x44, 0x30, 0x30, 0x31]),
        ],
        "handler": libarchive_handler,
    },
    "application/x-7z-compressed": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0x37, 0x7A, 0xBC, 0xAF, 0x27, 0x1C]),
        ],
        "handler": libarchive_handler,
    },
    "application/x-cpio": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0xC7, 0x71]),  # 070707 in octal (Little Endian).
            bytearray([0x71, 0xC7]),  # 070707 in octal (Big Endian).
            bytearray([0x30, 0x37, 0x30, 0x37, 0x30, 0x31]),  # "070701"
            bytearray([0x30, 0x37, 0x30, 0x37, 0x30, 0x32]),  # "070702"
            bytearray([0x30, 0x37, 0x30, 0x37, 0x30, 0x37]),  # "070707"
        ],
        "handler": libarchive_handler,
    },
    "application/x-xar": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0x78, 0x61, 0x72, 0x21]),
        ],
        "handler": xar_handler,
    },
    "application/vnd.ms-cab-compressed": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0x4D, 0x53, 0x43, 0x46]),
        ],
        "handler": libarchive_handler,
    },
    "application/x-archive": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0x21, 0x3C, 0x61, 0x72, 0x63, 0x68, 0x3E]),
        ],
        "handler": libarchive_handler,
    },
    "application/vnd.rar": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0x52, 0x61, 0x72, 0x21, 0x1A, 0x07]),
        ],
        "handler": libarchive_handler,
    },
    "application/zstd": {
        "weight": 1,
        "offset": 0,
        "magic": [
            bytearray([0x28, 0xB5, 0x2F, 0xFD]),
        ],
        "handler": zstd_handler,
    },
    "application/x-apple-diskimage": {
        "weight": 2,  # "container" formats are weighted higher.
        "offset": -512,
        "magic": [
            bytearray([0x6B, 0x6F, 0x6C, 0x79]),
        ],
        "handler": dmg_handler,
    },
}



"""Tests the STACS SARIF output module."""

import unittest

import stacs.scan


class STACSOutputSARIFTestCase(unittest.TestCase):
    """Tests the STACS SARIF output module."""

    def setUp(self):
        """Ensure the application is setup for testing."""
        pass

    def tearDown(self):
        """Ensure everything is torn down between tests."""
        pass

    def test_add_artifact(self):
        """Ensure that artifact entries are deduplicated by their full path."""
        findings = [
            stacs.scan.model.finding.Entry(
                path="/tmp/rootfs/etc/passwd",
                md5="b39bfc0e26a30024c76e4dcb8a1eae87",
            ),
            stacs.scan.model.finding.Entry(
                path="/tmp/rootfs/etc/passwd",
                md5="b39bfc0e26a30024c76e4dcb8a1eae87",
            ),
            stacs.scan.model.finding.Entry(
                path="/tmp/rootfs/a.tar.gz!a.tar!cred",
                md5="bf072e9119077b4e76437a93986787ef",
            ),
            stacs.scan.model.finding.Entry(
                path="/tmp/rootfs/a.tar.gz!a.tar!b_cred",
                md5="30cf3d7d133b08543cb6c8933c29dfd7",
            ),
            stacs.scan.model.finding.Entry(
                path="/tmp/rootfs/b.tar.gz!b_cred",
                md5="57b8d745384127342f95660d97e1c9c2",
            ),
            stacs.scan.model.finding.Entry(
                path="/tmp/rootfs/b.tar.gz!a.tar!cred",
                md5="787c9a8e2148e711f6e9f44696cf341f",
            ),
            stacs.scan.model.finding.Entry(
                path="/tmp/rootfs/a.tar.gz!a.tar!b.tar.gz!b.tar!pass",
                md5="d2a33790e5bf28b33cdbf61722a06989",
            ),
        ]

        # Ensure we get the expected number of artifacts in the artifacts list.
        artifacts = []
        for finding in findings:
            _, artifacts = stacs.scan.output.sarif.add_artifact(
                "/tmp/rootfs/", finding, artifacts
            )

        # Ensure findings are unfurled into the expected number of unique artifacts.
        self.assertEqual(len(artifacts), 12)



"""Setup required for pybind11 built native code only."""

import os
import platform
import subprocess
from typing import List

from pybind11.setup_helpers import Pybind11Extension
from setuptools import setup

ext_modules = [
    Pybind11Extension(
        "stacs.native.archive",
        ["stacs/native/archive/src/archive.cpp"],
        libraries=["archive"],
    ),
]


def run(command: List[str]):
    """Run a command, returning the output as a string or an exception on failure."""
    result = subprocess.run(command, capture_output=True, check=True)
    return str(result.stdout, "utf-8").strip()


# macOS requires a bit of special handling to ensure that the - likely - brew installed
# libarchive is discoverable. The macOS built-in libarchive is no good, as it's too
# old.
if platform.system() == "Darwin":
    libarchive = run(["brew", "--cellar", "libarchive"])
    libarchive_headers = run(["find", libarchive, "-name", "include", "-type", "d"])
    libarchive_pkgconfig = run(["find", libarchive, "-name", "pkgconfig", "-type", "d"])

    # Setup the environment for the build.
    os.environ["LDFLAGS"] = f"-L{libarchive_headers}"
    os.environ["PKG_CONFIG"] = libarchive_pkgconfig
    os.environ["CPPFLAGS"] = " ".join(
        [
            os.environ.get("CPPFLAGS", ""),
            "-std=c++11",
            f"-I{libarchive_headers}",
        ]
    )

setup(ext_modules=ext_modules, packages=[])



"""STACS - Static Token And Credential Scanner.

SPDX-License-Identifier: BSD-3-Clause
"""

from stacs.scan import __about__  # noqa:F401
from stacs.scan import constants  # noqa:F401
from stacs.scan import filter  # noqa:F401
from stacs.scan import helper  # noqa:F401
from stacs.scan import loader  # noqa:F401
from stacs.scan import model  # noqa:F401
from stacs.scan import output  # noqa:F401
from stacs.scan import scanner  # noqa:F401



"""Defines the primary STACS CLI entrypoint.

SPDX-License-Identifier: BSD-3-Clause
"""

import logging
import os
import shutil
import sys
import time
from types import TracebackType
from typing import Callable, List

import click
import stacs


def unlink_error(function: Callable, path: str, exc_info: TracebackType):
    """Provides a mechanism to better handle failures to delete files after a run.

    Currently, this just logs out. In future we should look to fix the permissions on
    the path / parent and call func(path) to attempt the deletion again. However, we'll
    need to ensure that path is actually part of the cache directory. So for now, we
    log.
    """
    logger = logging.getLogger("stacs")
    logger.warning(f"Unable to remove {path}")


@click.command()
@click.version_option()
@click.option(
    "--debug",
    is_flag=True,
    help="Increase verbosity of logs for debugging",
)
@click.option(
    "--pretty",
    help="Display outputs in a human-readable tree, rather than SARIF.",
    is_flag=True,
)
@click.option(
    "--threads",
    help="The number of threads to use when processing files",
    default=10,
)
@click.option(
    "--rule-pack",
    help="The path to the rule pack to load.",
    default="~/.stacs/pack.json",
)
@click.option(
    "--ignore-list",
    help="The path to the ignore list to load (if required).",
)
@click.option(
    "--skip-unprocessable",
    help="Skip unprocessable / corrupt archives with a warning.",
    is_flag=True,
)
@click.option(
    "--cache-directory",
    help="The path to use as a cache - used when unpacking archives.",
    default=stacs.scan.constants.CACHE_DIRECTORY,
)
@click.argument("paths", nargs=-1, required=True)
def main(
    debug: bool,
    pretty: bool,
    threads: int,
    rule_pack: str,
    ignore_list: str,
    skip_unprocessable: bool,
    cache_directory: str,
    paths: List[str],
) -> None:
    """STACS - Static Token And Credential Scanner."""
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.INFO,
        format="%(asctime)s - %(process)d - [%(levelname)s] %(message)s",
    )
    logger = logging.getLogger("stacs")
    logger.info(f"STACS running with {threads} threads")

    # Licenses.
    for project, urls in stacs.scan.constants.EXTERNAL_LICENSES.items():
        logger.info(f"STACS uses {project} (licenses may be found at {' '.join(urls)})")

    # Load the rule pack.
    logger.info(f"Attempting to load rule pack from {rule_pack}")
    try:
        pack = stacs.scan.model.pack.from_file(rule_pack)
    except stacs.scan.exceptions.STACSException as err:
        logger.error(f"Unable to load rule pack: {err}")
        sys.exit(-1)

    # Load the ignore list.
    ignored = []
    if ignore_list:
        logger.info(f"Attempting to load ignore list from {ignore_list}")
        try:
            ignored = stacs.scan.model.ignore_list.from_file(ignore_list)
            logger.debug(f"Loaded {len(ignored.ignore)} suppressions from ignore list.")
        except stacs.scan.exceptions.STACSException as err:
            logger.error(f"Unable to load ignore list: {err}")
            sys.exit(-1)

    # Append a timestamp to the cache directory to reduce the chance of collisions.
    cache_directory = os.path.join(cache_directory, str(int(time.time_ns() / 1000)))
    try:
        os.mkdir(cache_directory)
        logger.info(f"Using cache directory at {cache_directory}")
    except OSError as err:
        logger.error(f"Unable to create cache directory at {cache_directory}: {err}")
        sys.exit(-2)

    # Generate a list of candidate files to scan.
    targets = []

    for path in paths:
        path = os.path.abspath(os.path.expanduser(path))
        logger.info(f"Attempting to get a list of files to scan from {path}")
        try:
            targets.extend(
                stacs.scan.loader.filepath.finder(
                    path,
                    cache_directory,
                    skip_on_corrupt=skip_unprocessable,
                    workers=threads,
                )
            )
        except stacs.scan.exceptions.STACSException as err:
            logger.error(f"Unable to generate file list: {err}")
            sys.exit(-2)

    # Submit files for analysis.
    logger.info(f"Found {len(targets)} files for analysis")

    findings = []
    for scanner in stacs.scan.scanner.__all__:
        try:
            findings.extend(
                getattr(stacs.scan.scanner, scanner).run(targets, pack, workers=threads)
            )
        except stacs.scan.exceptions.InvalidFormatException as err:
            logger.error(f"Unable to load a rule in scanner {scanner}: {err}")
            continue

    # Filter findings by allow list.
    if ignored:
        findings = stacs.scan.filter.ignore_list.process(findings, ignored)

    # Clean-up cache directory.
    shutil.rmtree(cache_directory, onerror=unlink_error)

    # Determine the correct exit status based on whether there were unsuppressed
    # findings.
    exit_code = 0

    for finding in findings:
        if not finding.ignore:
            exit_code = stacs.scan.constants.EXIT_CODE_UNSUPPRESSED

    # Pretty print, if requested.
    if pretty:
        logger.info("Generating 'pretty' output from findings")
        stacs.scan.output.pretty.render(findings, pack)
        sys.exit(exit_code)

    # Default to SARIF output to STDOUT.
    logger.info("Generating SARIF from findings")
    try:
        sarif = stacs.scan.output.sarif.render(path, findings, pack)
    except stacs.scan.exceptions.STACSException as err:
        logger.error(f"Unable to generate SARIF: {err}")
        sys.exit(-3)

    # TODO: Add file output as an option.
    logger.info(f"Found {len(findings)} findings")
    print(sarif)



"""Define helpers commonly used throughout STACS.

SPDX-License-Identifier: BSD-3-Clause
"""
from typing import List

import colorama
from stacs.scan.constants import ARCHIVE_FILE_SEPARATOR
from stacs.scan.exceptions import NoParentException


def generate_virtual_path(
    finding: "Finding",  # noqa: F821
    artifacts: "List[Artifact]",  # noqa: F821
):
    """Generate a virtual path for an input file."""
    virtual_path = finding.filepath

    try:
        parent = artifacts[finding.artifact].parent

        while True:
            name = artifacts[parent].filepath
            virtual_path = f"{name}{ARCHIVE_FILE_SEPARATOR}{virtual_path}"

            parent = artifacts[parent].parent
    except NoParentException:
        return virtual_path


def printi(string, indent: int = 4, prefix: str = None):
    """Super janky wrapper to print something indented."""
    for line in string.splitlines():
        if prefix:
            print(f"{prefix}", end="")

        print(f"{' ' * indent}" + line)


def banner(version: str) -> str:
    """Returns a STACS console banner."""
    banner = colorama.Fore.BLUE
    banner += rf"""
    ______________   ___________
   / ___/_  __/   | / ____/ ___/
   \__ \ / / / /| |/ /    \__ \
  ___/ // / / ___ / /___ ___/ /
 /____//_/ /_/  |_\____//____/

       STACS version {version}
    """
    return banner
